class
	AST_FUNCTION_CALL

inherit
	AST_EXPRESSION

create
	make,
	make_from_antlr,
	make_copy

feature {NONE}

	make (a_identifier: STRING; a_parameters: ARRAY [AST_EXPRESSION]; a_type_binding: AST_TYPE; a_declaration_binding: AST_FUNCTION_DECLARATION)
		require
			a_identifier /= Void
			Bindings_active.item implies a_type_binding /= Void
			Bindings_active.item implies a_declaration_binding /= Void
		local
			i: INTEGER
		do
			identifier_str := a_identifier
			parameters := create {TWO_WAY_LIST [AST_EXPRESSION]}.make
			if a_parameters /= Void then
				from
					i := 1
				until
					i > a_parameters.count
				loop
					parameters.extend (a_parameters[i])
					a_parameters[i].parent := Current
					i := i + 1
				end
			end
			type_binding := a_type_binding
			declaration_binding := a_declaration_binding
		end

	make_from_antlr (node: ANTLR_AST_NODE; init_parent: AST_NODE)
		require
			node /= Void and then node.type = Ast_function_call
		do
			parent := init_parent

			node.children.start
			if node.children.item.type = Ast_expression then
				identifier_expr := process_expression (node.children.item, Current)
			else
				identifier_str := node.children.item.token_text
			end
			node.children.forth
			from
				parameters := create {TWO_WAY_LIST [AST_EXPRESSION]}.make
			until
				node.children.after
			loop
				parameters.extend (process_expression (node.children.item, Current))
				node.children.forth
			end
		end

feature {ANY}

	identifier_str: STRING

	identifier_expr: AST_EXPRESSION

	parameters: LIST [AST_EXPRESSION]

feature {ANY} -- bindings

	type_binding: AST_TYPE assign set_type_binding

	set_type_binding (a_type_binding: AST_TYPE)
		require
			a_type_binding /= Void
		do
			type_binding := a_type_binding
		end

	declaration_binding: AST_FUNCTION_DECLARATION assign set_declaration_binding

	set_declaration_binding (a_declaration: AST_FUNCTION_DECLARATION)
		require
			a_declaration /= Void
		do
			declaration_binding := a_declaration
		end

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_function_call (Current)
			if visit_children then
				visit_node (identifier_expr, v)
				visit_list (parameters, v)
			end
			v.end_visit_function_call (Current)
			v.post_visit (Current)
		end

	ast_twin: like Current
		do
			create Result.make_copy (Current)
		end

	make_copy (other: like Current)
		do
			if other.identifier_str /= Void then
				identifier_str := other.identifier_str.twin
			else
				identifier_expr := other.identifier_expr.ast_twin
				identifier_expr.parent := Current
			end
			parameters := create {TWO_WAY_LIST [AST_EXPRESSION]}.make
			across other.parameters as c loop
				parameters.extend (c.item.ast_twin)
				parameters.last.parent := Current
			end
			type_binding := other.type_binding
			declaration_binding := other.declaration_binding
		end

invariant
	(identifier_str = Void) /= (identifier_expr = Void)
	parameters /= Void
	(Bindings_active.item and identifier_str /= Void) implies declaration_binding /= Void

end
