class
	OO_FUNCTIONS_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	NAME_UTILITIES

feature {ANY} -- TRANSFORMATION_STEP

	Heuristic_struct_pointer_arg,
	Heuristic_source_file_static,
	Heuristic_loose_coupling: INTEGER_8 = unique

	process_impl: BOOLEAN
		do
			-- first iteration
			analyze_source_file
			print_statistics (1)

			if not transformer.project_settings.disable_oo_reengineering then
				-- second interation
				from
				until
					not analyze_struct_pointer_args
				loop
				end
				print_statistics (2)

--				-- third iteration?
--				analyze_global_struct_pointer
--				print_statistics (3)

				-- third iteration
				from
				until
					not analyze_coupling
				loop
				end
				print_statistics (3)

				update_declarations
			end

			update_accesses

			Result := True
		end

	description: STRING
		do
			Result := "Assigning functions to classes"
		end

feature {NONE}

	analyze_source_file
		local
			l_classes: HASH_TABLE [AST_STRUCT_DECLARATION, STRING_8]
			l_class_name: STRING_8
		do
			create l_classes.make (16)

			across transformer.tree.declarations as c loop
				if attached {AST_EIFFEL_FEATURE} c.item as ef then
					l_class_name := eiffel_class_name (ef.source_info.source_file)
					if ef.source_info.is_project_source then
						l_class_name := "P_" + l_class_name
					else
						l_class_name := "S_" + l_class_name
					end

					if not l_classes.has (l_class_name) then
						l_classes[l_class_name] := create {AST_STRUCT_DECLARATION}.make (False, l_class_name)
						l_classes[l_class_name].source_info := create {SOURCE_INFO}.make_oo (ef.source_info.is_project_source)
					end
					l_classes[l_class_name].add_static_feature (ef)

					ef.oo_class_assignment_data := [Heuristic_source_file_static, Void, 0, False]
				end
			end

			across l_classes as c loop
				c.item.parent := transformer.tree
				transformer.tree.declarations.extend (c.item)
			end
		end

	analyze_struct_pointer_args: BOOLEAN
			-- Result: True if a feature has been changed
		local
			l_done: BOOLEAN
		do
			-- stop after the first struct pointer; only take it if it can be a member
			across transformer.tree.declarations as c loop
				if
					attached {AST_FUNCTION_DECLARATION} c.item as fd and then
					fd.source_info.is_project_source and then
					fd.oo_class_assignment_data.applied_heuristic = Heuristic_source_file_static
				then
					l_done := False
					across
						fd.parameter_list.parameters as pc
					until
						l_done
					loop
						if
							attached {AST_VARIABLE_DECLARATION} pc.item as vd and then
							attached {AST_POINTER_TYPE} vd.type as pt and then
							attached {AST_STRUCT_TYPE} pt.type as st and then
							attached {AST_STRUCT_DECLARATION} st.declaration_binding as sd and then
							sd.source_info.is_project_source
						then
							l_done := True
							if is_extractable (fd, vd) then
								fd.oo_class.remove_member_feature (fd)
								sd.add_feature (fd)
								fd.oo_class_assignment_data := [Heuristic_struct_pointer_arg, vd, pc.target_index, False]

								Result := True
							end
						end
					end
				end
			end
		end

	analyze_global_struct_pointer
		local
			l_global_struct_pointers: TWO_WAY_LIST [TUPLE [var: AST_VARIABLE_DECLARATION; sd: AST_STRUCT_DECLARATION]]
			l_project_name: STRING_8
		do
			create l_global_struct_pointers.make
			across transformer.tree.declarations as c loop
				if
					attached {AST_VARIABLE_DECLARATION} c.item as vd and then
					attached {AST_POINTER_TYPE} vd.type as pt and then
					attached {AST_STRUCT_TYPE} pt.type as st and then
					attached st.declaration_binding as sd and then
					vd.source_info.is_project_source
				then
					l_global_struct_pointers.extend ([vd, sd])
				end
			end

			l_project_name := transformer.tree.source_file_name
			across transformer.tree.declarations as declaration_cursor loop
				if
					attached {AST_FUNCTION_DECLARATION} declaration_cursor.item as fd and then
					fd.source_info.is_project_source and then
					fd.oo_class_assignment_data.applied_heuristic = Heuristic_source_file_static
				then
					across l_global_struct_pointers as global_cursor loop
						if is_extractable (fd, global_cursor.item.var) then
							Io.put_string ("<#>	")
							Io.put_string (l_project_name)
							Io.put_string ("	")
							Io.put_string (fd.identifier)
							Io.put_string ("	")
							Io.put_string (global_cursor.item.sd.identifier)
							Io.put_new_line
						end
					end
				end
			end
		end

	analyze_coupling: BOOLEAN
			-- Result: True if a feature has been changed
		local
			l_best_feature: AST_EIFFEL_FEATURE
			l_new_class: AST_STRUCT_DECLARATION
			l_max_calls: NATURAL_32
		do
			-- if a feature is only called from one other class, it's a candidate for moving it:
			-- - P_  -> P_  -> don't move
			-- - P_  -> PD_ -> move
			-- - PD_ -> P_  -> don't move
			-- - PD_ -> PD_ -> don't move
			across transformer.tree.declarations as c loop
				if
					attached {AST_EIFFEL_FEATURE} c.item as ef and then
					ef.source_info.is_project_source and then
					ef.oo_class_assignment_data.applied_heuristic = Heuristic_source_file_static and then
					not ef.oo_class_assignment_data.fixed and then
					ef.call_graph.in_calls.count.to_natural_32 > l_max_calls and then
					attached ef.call_graph.in_calls.linear_representation.first.ast_feature as an_incalling_feature and then
					an_incalling_feature.oo_class /= ef.oo_class and then
					across ef.call_graph.in_calls as c2 all c2.item.ast_feature.oo_class = an_incalling_feature.oo_class end and then
					an_incalling_feature.oo_class.identifier.starts_with ("PD_")
				then
					l_best_feature := ef
					l_new_class := an_incalling_feature.oo_class
					l_max_calls := ef.call_graph.in_calls.count.to_natural_32
				end
			end

			if l_best_feature /= Void then
				l_best_feature.oo_class.remove_member_feature (l_best_feature)
				l_new_class.add_feature (l_best_feature)
				l_best_feature.oo_class_assignment_data := [Heuristic_loose_coupling, Void, 0, False]

				across l_best_feature.call_graph.in_calls as calling_features_cursor loop
					calling_features_cursor.item.ast_feature.oo_class_assignment_data.fixed := True
				end

				Result := True
			end
		end

feature {NONE} -- extraction

	is_extractable (a_function: AST_FUNCTION_DECLARATION; a_argument: AST_VARIABLE_DECLARATION): BOOLEAN
			-- Can `a_argument' be removed from `a_function' and become `Current'?
			-- This requires that it is only dereferenced; in each possible path at least once.
		require
			a_function /= Void
			a_argument /= Void
			a_argument.parent.parent = a_function
			attached {AST_POINTER_TYPE} a_argument.type as pt and then attached {AST_STRUCT_TYPE} pt.type
		local
			l_dereferencing_info: HASH_TABLE [BOOLEAN, CFG_BASIC_BLOCK]
			l_invalid_access_found: CELL [BOOLEAN]
			l_cfg: CFG_BASIC_BLOCK
		do
			create l_dereferencing_info.make (16)
			create l_invalid_access_found.put (False)

			collect_dereferencing_info_rec (a_argument, a_function.c_cfg, l_dereferencing_info, l_invalid_access_found)
			if not l_invalid_access_found.item then
				-- ignore simple argument checks in the extraction analysis
				from
					l_cfg := a_function.c_cfg
				invariant
					l_cfg /= Void
				until
					not (
							l_cfg.out_edges.count = 2 and then
							attached l_cfg.get_out_edge_by_label ({CFG_EDGE}.Label_true) as true_edge and then
							attached l_cfg.get_out_edge_by_label ({CFG_EDGE}.Label_false) as false_edge and then
							true_edge.to_block.out_edges.count = 0 and then
							true_edge.to_block.in_edges.count = 1 and then
							contains_argument_usage (l_cfg, a_function, a_argument)
						)
				loop
					l_cfg := false_edge.to_block
				end

				Result := is_extractable_rec (l_cfg, l_dereferencing_info) = Res_extractable
			end
		end

	collect_dereferencing_info_rec (a_node: AST_VARIABLE_DECLARATION; a_block: CFG_BASIC_BLOCK; a_dereferencing_info: HASH_TABLE [BOOLEAN, CFG_BASIC_BLOCK]; a_invalid_access_found: CELL [BOOLEAN])
		require
			a_node /= Void
			a_block /= Void
			a_dereferencing_info /= Void and then not a_dereferencing_info.has (a_block)
			a_invalid_access_found /= Void and then not a_invalid_access_found.item
		local
			l_status: TUPLE [is_dereferenced, invalid_access: BOOLEAN]
			l_visitor: AST_AGENT_VISITOR
		do
			l_status := [False, False]

			create l_visitor
			l_visitor.agents.end_visit_variable_expression := agent (ag_node: AST_VARIABLE_EXPRESSION; ag_decl: AST_VARIABLE_DECLARATION; ag_status: TUPLE [is_dereferenced, invalid_access: BOOLEAN])
				do
					if ag_node.declaration_binding = ag_decl then
						check_dereferencing (ag_node, ag_status)
					end
				end (?, a_node, l_status)
			across
				a_block.nodes as c
			until
				l_status.invalid_access
			loop
				c.item.accept (l_visitor)
			end

			a_dereferencing_info[a_block] := l_status.is_dereferenced
			if l_status.invalid_access then
				a_invalid_access_found.put (True)
			else
				across
					a_block.out_edges as c
				until
					a_invalid_access_found.item
				loop
					if not a_dereferencing_info.has (c.item.to_block) then
						collect_dereferencing_info_rec (a_node, c.item.to_block, a_dereferencing_info, a_invalid_access_found)
					end
				end
			end
		end

	check_dereferencing (a_node: AST_EXPRESSION; a_status: TUPLE [is_dereferenced, invalid_access: BOOLEAN])
		require
			a_node /= Void
			a_status /= Void
		local
			l_rec_status: like a_status
		do
			-- `p->name'
			if
				attached {AST_POSTFIX_EXPRESSION} a_node.non_parentheses_parent as pe and then
				pe.type = C_op_arrow
				-- array accesses (e.g. `p[...]') are considered invalid (even `p[0]') assuming that the argument is a real array
			then
				a_status.is_dereferenced := True

			-- `*p'
			elseif
				attached {AST_UNARY_EXPRESSION} a_node.non_parentheses_parent as ue and then
				ue.type = C_op_mult
			then
				if
					attached {AST_ASSIGNMENT_EXPRESSION} ue.non_parentheses_parent as ae and then
					ae.target = ue.parentheses_parent
				then
					-- `*p = ...' is not supported
					a_status.invalid_access := True
				else
					a_status.is_dereferenced := True
				end

			-- `foo(p)' where `p' will be extracted
			elseif
				attached {AST_FUNCTION_CALL} a_node.non_parentheses_parent as function_call and then
				attached function_call.parameters.index_of (a_node.parentheses_parent, 1) as argument_index and then
				argument_index > 0 and then
				attached function_call.declaration_binding as called_function and then
				called_function.oo_class_assignment_data /= Void and then
				called_function.oo_class_assignment_data.removed_parameter_index = argument_index
			then
				a_status.is_dereferenced := True

			-- `create {...}.make_cast (p)'
			elseif
				attached {AST_EIFFEL_EXPRESSION} a_node.non_parentheses_parent as ee and then
				attached ee.previous_string (a_node.parentheses_parent) as previous_string and then
				previous_string.is_equal ("}.make_cast (")
			then
				l_rec_status := [False, False]
				check_dereferencing (ee, l_rec_status)
				if l_rec_status.is_dereferenced then
					a_status.is_dereferenced := True
				end
			end
		end

	contains_argument_usage (a_block: CFG_BASIC_BLOCK; a_function: AST_FUNCTION_DECLARATION; a_argument: AST_VARIABLE_DECLARATION): BOOLEAN
			-- Does `a_block' reference at least one argument of `a_function' but not `a_argument'?
		require
			a_block /= Void
			a_function /= Void
			a_argument /= Void
		local
			l_status: TUPLE [uses_argument, uses_forbidden_argument: BOOLEAN]
			l_visitor: AST_AGENT_VISITOR
		do
			l_status := [False, False]

			create l_visitor
			l_visitor.agents.end_visit_variable_expression := agent (ag_node: AST_VARIABLE_EXPRESSION; ag_function: AST_FUNCTION_DECLARATION; ag_argument: AST_VARIABLE_DECLARATION; ag_status: TUPLE [uses_argument, uses_forbidden_argument: BOOLEAN])
				do
					if ag_node.declaration_binding = ag_argument then
						ag_status.uses_forbidden_argument := True
					else
						ag_status.uses_argument := across ag_function.parameter_list.parameters as c some
								ag_node.declaration_binding = c.item
							end
					end
				end (?, a_function, a_argument, l_status)
			across a_block.nodes as c loop
				c.item.accept (l_visitor)
			end

			Result := l_status.uses_argument and not l_status.uses_forbidden_argument
		end

	Res_extractable,
	Res_not_extractable,
	Res_recursion_hit: INTEGER_8 = unique

	is_extractable_rec (a_block: CFG_BASIC_BLOCK; a_dereferencing_info: HASH_TABLE [BOOLEAN, CFG_BASIC_BLOCK]): INTEGER_8
		require
			a_block /= Void
			a_dereferencing_info /= Void
		local
			l_is_dereferenced: BOOLEAN
			l_has_extractable, l_has_not_extractable, l_has_recursion_hit: BOOLEAN
		do
			if a_dereferencing_info.has (a_block) then
				l_is_dereferenced := a_dereferencing_info[a_block]
				a_dereferencing_info.remove (a_block)

				if l_is_dereferenced then
					Result := Res_extractable
				else
					across
						a_block.out_edges as c
					until
						l_has_not_extractable
					loop
						inspect is_extractable_rec (c.item.to_block, a_dereferencing_info)
						when Res_extractable then     l_has_extractable := True
						when Res_not_extractable then l_has_not_extractable := True
						when Res_recursion_hit then   l_has_recursion_hit := True
						end
					end
					    if l_has_not_extractable then Result := Res_not_extractable
					elseif l_has_extractable then     Result := Res_extractable
					elseif l_has_recursion_hit then   Result := Res_recursion_hit
					else                              Result := Res_not_extractable
					end
				end
			else
				Result := Res_recursion_hit
			end
		ensure
			Result = Res_extractable or Result = Res_not_extractable or Result = Res_recursion_hit
		end

feature {NONE} -- update

	update_declarations
			-- Extracting the argument from `a_function'. E.g.:
			-- `set_age(struct person *p, int age) { p->age = age; }' -> `set_age(int age) { Current.age = age; }'
		local
			l_data: CELL [AST_VARIABLE_DECLARATION]
			l_visitor: AST_AGENT_VISITOR
		do
			create l_data.put (Void)
			create l_visitor
			l_visitor.agents.end_visit_variable_expression := agent (ag_node: AST_VARIABLE_EXPRESSION; ag_var_decl: CELL [AST_VARIABLE_DECLARATION])
				local
					l_pe: AST_POSTFIX_EXPRESSION  -- `p->name' | `p.name'
					l_ue: AST_UNARY_EXPRESSION    -- `*p'
					l_ve: AST_VARIABLE_EXPRESSION -- `p'
				do
					if ag_node.declaration_binding = ag_var_decl.item then
						if
							attached {AST_POSTFIX_EXPRESSION} ag_node.non_parentheses_parent as pe and then
							(pe.type = C_op_arrow or pe.type = C_op_point)
						then
							l_pe := pe
						elseif
							attached {AST_UNARY_EXPRESSION} ag_node.non_parentheses_parent as ue and then
							ue.type = C_op_mult
						then
							l_ue := ue
						else
							l_ve := ag_node
						end

						-- `p->name' -> `name' | `p.name' -> `name'
						if
							l_pe /= Void and then
							not (attached {AST_ASSIGNMENT_EXPRESSION} l_pe.non_parentheses_parent as ae and then ae.target = l_pe.parentheses_parent)
						then
							if attached {AST_VARIABLE_DECLARATION} l_pe.declaration_binding as vd then
								replace_node (l_pe, create {AST_VARIABLE_EXPRESSION}.make (vd))
							else
								-- TODO: Use the binding; the field might be renamed later.
								replace_node (l_pe, create {AST_EIFFEL_EXPRESSION}.make (l_pe.declaration_binding.identifier, l_pe.type_binding))
							end

						-- `(*p).c_pointer' -> `c_pointer'
						-- `(*p).twin' -> `twin'
						elseif
							l_ue /= Void and then
							attached {AST_EIFFEL_EXPRESSION} l_ue.non_parentheses_parent as ee and then
							attached l_ue.parentheses_parent as pp and then
							attached ee.next_string (pp) as s and then
							(
								s.starts_with (".c_pointer") or
								s.starts_with (".twin")
							)
						then
							s.remove_head (1)
							delete_node (pp)

						-- `p.c_pointer' -> `c_pointer'
						-- `p.twin' -> `twin'
						elseif
							l_ve /= Void and then
							attached {AST_EIFFEL_EXPRESSION} l_ve.non_parentheses_parent as ee and then
							attached l_ve.parentheses_parent as pp and then
							attached ee.next_string (pp) as s and then
							(
								s.starts_with (".c_pointer") or
								s.starts_with (".twin")
							)
						then
							s.remove_head (1)
							delete_node (pp)

						-- `create {...}.make_cast (p)' -> `create {...}.make_shared (c_pointer)'
						elseif
							l_ve /= Void and then
							attached {AST_EIFFEL_EXPRESSION} l_ve.non_parentheses_parent as ee and then
							attached l_ve.parentheses_parent as pp and then
							attached ee.previous_string (pp) as previous_string and then
							previous_string.is_equal ("}.make_cast (") and then
							attached ee.next_string (pp) as next_string and then
							next_string.starts_with (")")
						then
							previous_string.make_from_string ("}.make_shared (c_pointer")
							delete_node (pp)

						-- `p.is_equal (foo)' -> `(c_pointer = foo.c_pointer)'
						elseif
							l_ve /= Void and then
							attached {AST_EIFFEL_EXPRESSION} l_ve.non_parentheses_parent as ee and then
							attached l_ve.parentheses_parent as pp and then
							attached ee.next_string (pp) as next_string and then
							next_string.is_equal (".is_equal (") and then
							attached ee.next_node (next_string) as other_var and then
							attached ee.next_string (other_var) as closing_parentheses and then
							closing_parentheses.is_equal (")")
						then
							next_string.make_from_string ("(c_pointer = (")
							closing_parentheses.make_from_string (").c_pointer)")
							delete_node (pp)

						-- `foo.is_equal (p)' -> `(foo.c_pointer = c_pointer)'
						elseif
							l_ve /= Void and then
							attached {AST_EIFFEL_EXPRESSION} l_ve.non_parentheses_parent as ee and then
							attached l_ve.parentheses_parent as pp and then
							attached ee.previous_string (pp) as previous_string and then
							previous_string.is_equal (".is_equal (") and then
							attached ee.previous_node (previous_string) as other_var
						then
							previous_string.make_from_string (".c_pointer = c_pointer")
							ee.insert_before (other_var, "(")
							delete_node (pp)

						-- `p.is_void' -> `(c_pointer = default_pointer)'
						elseif
							l_ve /= Void and then
							attached {AST_EIFFEL_EXPRESSION} l_ve.non_parentheses_parent as ee and then
							attached l_ve.parentheses_parent as pp and then
							attached ee.next_string (pp) as next_string and then
							next_string.is_equal (".is_void")
						then
							next_string.make_from_string ("(c_pointer = default_pointer)")
							delete_node (pp)

						-- `p->foo' -> `Current.foo'
						elseif l_pe /= Void then
							replace_node (l_pe, create {AST_POSTFIX_EXPRESSION}.make_point (
									create {AST_EIFFEL_EXPRESSION}.make ("Current", l_pe.main_expression.type_binding),
									l_pe.identifier,
									l_pe.type_binding,
									l_pe.declaration_binding
								))

						-- `*p' -> `Current'
						elseif l_ue /= Void then
							replace_node (l_ue, create {AST_EIFFEL_EXPRESSION}.make ("Current", l_ue.type_binding))

						-- `p' -> `(create {CE_POINTER [FOO]}.make_shared (c_pointer)'
						elseif l_ve /= Void then
							replace_node (l_ve, create {AST_EIFFEL_EXPRESSION}.make_mixed (<<
									"(create {",
									l_ve.type_binding.ast_twin,
									"}.make_shared (c_pointer))"
								>>, l_ve.type_binding))
						end
					end
				end (?, l_data)

			across transformer.tree.declarations as c loop
				if
					attached {AST_FUNCTION_DECLARATION} c.item as fd and then
					fd.source_info.is_project_source and then
					fd.oo_class_assignment_data.applied_heuristic = Heuristic_struct_pointer_arg
				then
					l_data.put (fd.oo_class_assignment_data.removed_parameter)
					fd.accept (l_visitor)
					delete_node (fd.oo_class_assignment_data.removed_parameter)
				end
			end
		end

	update_accesses
			-- Adapting usages of function calls and variable accesses. E.g.:
			-- `static_example(...)'          -> `(create {BAR}).static_example(...)'
			-- `struct_pointer_example(expr)' -> `*(expr).struct_pointer_example()'
		local
			l_visitor: AST_AGENT_VISITOR
		do
			create l_visitor
			l_visitor.agents.end_visit_function_call := agent (ag_node: AST_FUNCTION_CALL)
				local
					l_target_expr: AST_EXPRESSION
				do
					if
						attached ag_node.declaration_binding as called_function and then
						attached called_function.oo_class as called_class and then
						called_class.source_info.is_project_source and then
						attached ({AST_EIFFEL_FEATURE} #? ag_node.parent_of_type ({AST_EIFFEL_FEATURE})).oo_class as calling_class
					then
						if called_function.oo_class_assignment_data.applied_heuristic = Heuristic_struct_pointer_arg then
							l_target_expr := ag_node.parameters.i_th (called_function.oo_class_assignment_data.removed_parameter_index)
							delete_node (l_target_expr)

							if
								attached {AST_EIFFEL_EXPRESSION} l_target_expr as ee and then
								ee.mixed.count = 1 and then
								ee.is_string (1, "Current")
							then
								-- don't use "Current" as target
							else
								l_target_expr := create {AST_PARENTHESIZED_EXPRESSION}.make (l_target_expr.ast_twin)
								if attached {AST_POINTER_TYPE} l_target_expr.type_binding or attached {AST_ARRAY_TYPE} l_target_expr.type_binding then
									l_target_expr := create {AST_UNARY_EXPRESSION}.make (C_op_mult, l_target_expr)
								end

								replace_node (ag_node, create {AST_EIFFEL_EXPRESSION}.make_mixed (<<
										l_target_expr,
										".",
										ag_node.ast_twin
									>>, ag_node.type_binding))
							end
						elseif calling_class /= called_class then
							replace_node (ag_node, create {AST_EIFFEL_EXPRESSION}.make_mixed (<<
									"(create {",
									create {AST_STRUCT_TYPE}.make (called_class.identifier, called_class.is_union, called_class),
									"_STATIC}).",
									ag_node.ast_twin
								>>, ag_node.type_binding))
						end
					end
				end
			l_visitor.agents.end_visit_variable_expression := agent (ag_node: AST_VARIABLE_EXPRESSION)
				do
					if
						attached ag_node.declaration_binding as called_variable and then
						attached {AST_TRANSLATION_UNIT} called_variable.parent and then
						attached called_variable.oo_class as called_class and then
						called_class.source_info.is_project_source and then
						attached ({AST_EIFFEL_FEATURE} #? ag_node.parent_of_type ({AST_EIFFEL_FEATURE})).oo_class as calling_class and then
						called_class /= calling_class
					then
						replace_node (ag_node, create {AST_EIFFEL_EXPRESSION}.make_mixed (<<
								"(create {",
								create {AST_STRUCT_TYPE}.make (called_class.identifier, called_class.is_union, called_class),
								"_STATIC}).",
								ag_node.ast_twin
							>>, ag_node.type_binding))
					end
				end
			transformer.tree.accept (l_visitor)
		end

feature {NONE} -- statistic stuff

	print_statistics (a_finished_iteration: INTEGER_32)
		do
--			log ("----- OO_FUNCTIONS_TRANSFORMER, Iteration " + a_finished_iteration.out + " -----%N")
--			print_assignment_statistics (a_finished_iteration)
--			print_class_member_count (a_finished_iteration)
--			print_coupling_statistics (a_finished_iteration)
--			print_struct_arg_statistics (a_finished_iteration)
--			print_member_class_assignment (a_finished_iteration)
--			print_multiple_struct_pointer_args (a_finished_iteration)
		end

	print_assignment_statistics (a_finished_iteration: INTEGER_32)
		local
			l_source_file_static_count_routines, l_struct_pointer_arg_count_routines, l_loose_coupling_count_routines: NATURAL_32
			l_source_file_static_count_attributes, l_struct_pointer_arg_count_attributes, l_loose_coupling_count_attributes: NATURAL_32
			l_struct_field_count: NATURAL_64
		do
			across transformer.tree.declarations as declaration_cursor loop
				if
					attached {AST_EIFFEL_FEATURE} declaration_cursor.item as ef and then
					ef.source_info.is_project_source
				then
					if attached {AST_FUNCTION_DECLARATION} ef then
						inspect ef.oo_class_assignment_data.applied_heuristic
						when Heuristic_struct_pointer_arg then l_struct_pointer_arg_count_routines := l_struct_pointer_arg_count_routines + 1
						when Heuristic_loose_coupling then     l_loose_coupling_count_routines := l_loose_coupling_count_routines + 1
						when Heuristic_source_file_static then l_source_file_static_count_routines := l_source_file_static_count_routines + 1
						end
					else
						inspect ef.oo_class_assignment_data.applied_heuristic
						when Heuristic_struct_pointer_arg then l_struct_pointer_arg_count_attributes := l_struct_pointer_arg_count_attributes + 1
						when Heuristic_loose_coupling then     l_loose_coupling_count_attributes := l_loose_coupling_count_attributes + 1
						when Heuristic_source_file_static then l_source_file_static_count_attributes := l_source_file_static_count_attributes + 1
						end
					end
				elseif
					attached {AST_STRUCT_DECLARATION} declaration_cursor.item as sd and then
					sd.source_info.is_project_source
				then
					across sd.fields as field_cursor loop
						if field_cursor.item.identifier /= Void then
							l_struct_field_count := l_struct_field_count + 1
						end
					end
				end
			end
			log ("Assignment Statistics; Routines:")
			log ("- " + l_struct_pointer_arg_count_routines.out + " heuristic_struct_pointer_arg")
			log ("- " + l_loose_coupling_count_routines.out + " heuristic_loose_coupling")
			log ("- " + l_source_file_static_count_routines.out + " heuristic_source_file_static")
			log ("Assignment Statistics; Attributes:")
			log ("- " + l_struct_pointer_arg_count_attributes.out + " heuristic_struct_pointer_arg")
			log ("- " + l_loose_coupling_count_attributes.out + " heuristic_loose_coupling")
			log ("- " + l_source_file_static_count_attributes.out + " heuristic_source_file_static")
			log ("Struct/Union fields:")
			log ("- " + l_struct_field_count.out + " struct_field_count")
			log ("")
		end

	print_class_member_count (a_finished_iteration: INTEGER_32)
		local
			l_project_name: STRING_8
			l_member_count: NATURAL_64
			l_private_count: NATURAL_64
		do
			l_project_name := transformer.tree.source_file_name
			across transformer.tree.declarations as declaration_cursor loop
				if
					attached {AST_STRUCT_DECLARATION} declaration_cursor.item as sd and then
					sd.source_info.is_project_source
				then
					l_member_count := sd.oo_features.count.to_natural_64 + sd.oo_static_features.count.to_natural_64
					across sd.fields as c loop
						if c.item.identifier /= Void then
							l_member_count := l_member_count + 1
						end
					end

					l_private_count := 0
					across sd.oo_features as c loop
						if c.item.oo_class_assignment_data.applied_heuristic = Heuristic_loose_coupling then
							l_private_count := l_private_count + 1
						end
					end
					across sd.oo_static_features as c loop
						if c.item.oo_class_assignment_data.applied_heuristic = Heuristic_loose_coupling then
							l_private_count := l_private_count + 1
						end
					end

					Io.put_string ("##-->	iteration=")
					Io.put_integer_32 (a_finished_iteration)
					Io.put_string ("	")
					Io.put_string (l_project_name)
					Io.put_string ("	")
					Io.put_string (sd.identifier)
					Io.put_string ("	")
					Io.put_natural_64 (l_member_count)
					Io.put_string ("	")
					Io.put_natural_64 (l_private_count)
					Io.put_new_line
				end
			end
		end

	print_coupling_statistics (a_finished_iteration: INTEGER_32)
		local
			l_data: HASH_TABLE [TUPLE [self_calls, out_calls: NATURAL_64], AST_STRUCT_DECLARATION]
			l_visitor_data: TUPLE [current_var: AST_VARIABLE_DECLARATION; accessed_fields: HASH_SET [AST_STRUCT_FIELD]]
			l_visitor: AST_AGENT_VISITOR
			l_total_self_calls, l_total_out_calls: NATURAL_64
			l_project_name: STRING_8
		do
			create l_visitor_data
			create l_visitor
			l_visitor.agents.end_visit_variable_expression := agent (
						ag_node: AST_VARIABLE_EXPRESSION;
						ag_data: TUPLE [current_var: AST_VARIABLE_DECLARATION; accessed_fields: HASH_SET [AST_STRUCT_FIELD]]
					)
				do
					if ag_node.declaration_binding = ag_data.current_var then
						if
							attached {AST_POSTFIX_EXPRESSION} ag_node.non_parentheses_parent as pe and then
							pe.type = C_op_arrow and then
							attached pe.declaration_binding as accessed_field
						then
							-- `p->age'
							ag_data.accessed_fields.extend (accessed_field)
						elseif
							attached {AST_UNARY_EXPRESSION} ag_node.non_parentheses_parent as ue and then
							ue.type = C_op_mult and then
							attached {AST_POSTFIX_EXPRESSION} ue.non_parentheses_parent as pe and then
							pe.type = C_op_point and then
							attached pe.declaration_binding as accessed_field

						then
							-- `(*p).age'
							ag_data.accessed_fields.extend (accessed_field)
						end
					end
				end (?, l_visitor_data)

			create l_data.make (16)
			across transformer.tree.declarations as declaration_cursor loop
				if
					attached {AST_EIFFEL_FEATURE} declaration_cursor.item as calling and then
					calling.source_info.is_project_source
				then
					-- call graph; calling routines and globals
					across calling.call_graph.out_calls as out_call_cursor loop
						if
							attached out_call_cursor.item.ast_feature as called and then
							called.source_info.is_project_source
						then
							if not l_data.has (calling.oo_class) then
								l_data[calling.oo_class] := [{NATURAL_64} 0, {NATURAL_64} 0]
							end
							if calling.oo_class = called.oo_class then
								--log(calling.oo_class.identifier + ": " + calling.identifier + " -> " + called.identifier)
								l_data[calling.oo_class].self_calls := l_data[calling.oo_class].self_calls + 1
							else
								--log(calling.oo_class.identifier + "." + calling.identifier + " -> " + called.oo_class.identifier + "." + called.identifier)
								l_data[calling.oo_class].out_calls := l_data[calling.oo_class].out_calls + 1
							end
						end
					end
					-- O-O refactoring; accessing class attributres
					if attached calling.oo_class_assignment_data.removed_parameter as new_current then
						l_visitor_data.current_var := new_current
						l_visitor_data.accessed_fields := create {HASH_SET [AST_STRUCT_FIELD]}.make (16)
						calling.accept (l_visitor)

						if not l_visitor_data.accessed_fields.is_empty then
							--across l_visitor_data.accessed_fields as c loop
							--	log(calling.oo_class.identifier + ": " + calling.identifier + " -> " + c.item.identifier)
							--end
							if not l_data.has (calling.oo_class) then
								l_data[calling.oo_class] := [{NATURAL_64} 0, {NATURAL_64} 0]
							end
							l_data[calling.oo_class].self_calls := l_data[calling.oo_class].self_calls + l_visitor_data.accessed_fields.count.to_natural_64
						end
					end
				end
			end

			l_project_name := transformer.tree.source_file_name
			--Io.put_string ("### Coupling Statistics (project class_name out_calls in_calls):%N")
			if not l_data.is_empty then
				across l_data as c loop
					l_total_self_calls := l_total_self_calls + c.item.self_calls
					l_total_out_calls := l_total_out_calls + c.item.out_calls

					Io.put_string ("###-###-->	")
					Io.put_string (l_project_name)
					Io.put_string ("	")
					Io.put_string (c.key.identifier)
					Io.put_string ("	")
					Io.put_natural_64 (c.item.out_calls)
					Io.put_string ("	")
					Io.put_natural_64 (c.item.self_calls)
					Io.put_new_line
					--log ("--> " + c.key.identifier + ": " + c.item.self_calls.out + "/" + c.item.out_calls.out)
				end

				--log ("--> total (inside/outside): " + l_total_self_calls.out + "/" + l_total_out_calls.out)
			else
				log ("--> no calls")
			end
			log ("")
		end

	print_struct_arg_statistics (a_finished_iteration: INTEGER_32)
		local
			l_project_name: STRING_8
		do
			l_project_name := transformer.tree.source_file_name
			across transformer.tree.declarations as declaration_cursor loop
				if
					attached {AST_FUNCTION_DECLARATION} declaration_cursor.item as fd and then
					across fd.parameter_list.parameters as argument_cursor some
							(
								attached {AST_STRUCT_TYPE} argument_cursor.item or
								(attached {AST_VARIABLE_DECLARATION} argument_cursor.item as vd and then attached {AST_STRUCT_TYPE} vd.type)
							)
						end
				then
					log ("<#> " + l_project_name + "	" + fd.identifier + "	" + fd.c_code.split ('%N').first)
				end
			end
		end

	print_member_class_assignment (a_finished_iteration: INTEGER_32)
		local
			l_project_name: STRING_8
		do
			l_project_name := transformer.tree.source_file_name
			across transformer.tree.declarations as declaration_cursor loop
				if
					attached {AST_FUNCTION_DECLARATION} declaration_cursor.item as fd and then
					fd.oo_class_assignment_data.applied_heuristic = Heuristic_struct_pointer_arg
				then
					Io.put_string ("<#>	")
					Io.put_string (l_project_name)
					Io.put_string ("	")
					Io.put_string (fd.identifier)
					Io.put_string ("	")
					Io.put_string (fd.oo_class.identifier)
					Io.put_new_line
				end
			end
		end


	print_multiple_struct_pointer_args (a_finished_iteration: INTEGER_32)
		local
			l_project_name: STRING_8
			l_first: AST_VARIABLE_DECLARATION
		do
			l_project_name := transformer.tree.source_file_name
			across transformer.tree.declarations as c loop
				if
					attached {AST_FUNCTION_DECLARATION} c.item as fd and then
					fd.source_info.is_project_source
				then
					l_first := Void
					across
						fd.parameter_list.parameters as pc
					loop
						if
							attached {AST_VARIABLE_DECLARATION} pc.item as vd and then
							attached {AST_POINTER_TYPE} vd.type as pt and then
							attached {AST_STRUCT_TYPE} pt.type as st and then
							attached {AST_STRUCT_DECLARATION} st.declaration_binding as sd and then
							sd.source_info.is_project_source
						then
							if l_first = Void then
								l_first := vd
							elseif
								not vd.type.is_equal_type (l_first.type) and
								is_extractable (fd, vd)
							then
								Io.put_string ("<#>	")
								Io.put_string (l_project_name)
								Io.put_string ("	")
								Io.put_string (vd.eiffel_code)
								Io.put_string ("	(")
								Io.put_string (pc.target_index.out)
								Io.put_string (")	")
								Io.put_string (fd.eiffel_code.split ('%N')[1])
								Io.put_new_line
							end
						end
					end
				end
			end
		end

end
