class
	PRINTF_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			visit_translation_unit,
			end_visit_translation_unit,
			end_visit_function_declaration,
			end_visit_function_call,
			end_visit_unary_expression
		end

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		local
			l_retry: BOOLEAN
		do
			if not l_retry then
				transformer.tree.accept (Current)
				--log ("replaced " + replaced_printf.out + " out of " + total_printf.out + " printf statements")
				Result := True
			else
				log ((create {EXCEPTION_MANAGER}).last_exception.message)
				Result := False
			end
		rescue
			l_retry := True
			retry
		end

	description : STRING
		do
			Result := "Transforming %"printf%" statements"
		end

feature {NONE}

	printf_declaration, fprintf_declaration: AST_FUNCTION_DECLARATION

	extern_printf_needed, extern_fprintf_needed: BOOLEAN

	total_printf, replaced_printf: INTEGER

feature {ANY} -- AST_DEFAULT_VISITOR

	visit_translation_unit (node: AST_TRANSLATION_UNIT): BOOLEAN
		do
			extern_printf_needed := False
			printf_declaration := Void
			extern_fprintf_needed := False
			fprintf_declaration := Void
			Result := True
		end

	end_visit_translation_unit (node: AST_TRANSLATION_UNIT)
		do
			if not extern_printf_needed and printf_declaration /= Void then
				delete_node (printf_declaration)
			end
			extern_printf_needed := False
			printf_declaration := Void

			if not extern_fprintf_needed and fprintf_declaration /= Void then
				delete_node (fprintf_declaration)
			end
			extern_fprintf_needed := False
			fprintf_declaration := Void
		end

	end_visit_function_declaration (node: AST_FUNCTION_DECLARATION)
		do
			if
				node.is_extern and then
				node.identifier.is_equal ("printf")
			then
				printf_declaration := node
			elseif
				node.is_extern and then
				node.identifier.is_equal ("fprintf")
			then
				fprintf_declaration := node
			end
		end

	end_visit_function_call (node: AST_FUNCTION_CALL)
		local
			helper: PRINTF_HELPER
			eiffel_parameter: AST_EXPRESSION
			output: STRING
			consider: BOOLEAN
		do
			-- `printf("foo %d bar", an_int);'  ->  `Io.put_string ("foo " + (an_int).out + " bar")'
			create helper
			output := "Io"
			if
				attached {AST_FUNCTION_DECLARATION} node.declaration_binding as decl and then
				(decl.identifier.is_equal ("printf") or decl.identifier.is_equal ("fprintf")) and then
				decl.source_info.is_system_source and then
				not node.parameters.is_empty
			then
				node.parameters.start

				consider := decl.identifier.is_equal ("printf") or else
							(decl.identifier.is_equal ("fprintf") and then
							attached {AST_VARIABLE_EXPRESSION} node.parameters.first as expr and then
							(expr.identifier.is_equal ("stdout") or expr.identifier.is_equal ("stderr")))

				if consider then
					total_printf := total_printf + 1

					if attached {AST_EXPRESSION_STATEMENT} node.parent then
						if
							decl.identifier.is_equal ("printf") and then
							helper.is_printf_replaceable (node.parameters)
						then
							eiffel_parameter := helper.create_eiffel_parameters (node.parameters)
						elseif
							decl.identifier.is_equal ("fprintf") and then
							attached {AST_VARIABLE_EXPRESSION} node.parameters.item as expr and then
							(expr.identifier.is_equal ("stdout") or expr.identifier.is_equal ("stderr")) and then
							not node.parameters.islast
						then
							if expr.identifier.is_equal ("stderr") then
								output.append (".error")
							end
							node.parameters.forth

							if helper.is_printf_replaceable (node.parameters) then
								eiffel_parameter := helper.create_eiffel_parameters (node.parameters)
							end
						else
							-- not supported
							eiffel_parameter := Void
						end
					end
				end

				if eiffel_parameter /= Void then
					replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (<< output + ".put_string (", eiffel_parameter, ")" >>, create {AST_BASIC_TYPE}.make (Ast_bt_void)))
					replaced_printf := replaced_printf + 1
				else
					if decl.identifier.is_equal ("printf") then
						extern_printf_needed := True
					else
						extern_fprintf_needed := True
					end

					if consider then
						--log ("Not supported: %"" + node.c_code + "%"")
					end

					across
						node.parameters as p
					loop
						if attached {AST_EIFFEL_STRING_TYPE} p.item.type_binding then
							-- convert eiffel strings to ce_strings for external printf call
							replace_node (p.item, create {AST_EIFFEL_EXPRESSION}.make_mixed (<< "ce_string (", p.item.ast_twin, ")" >>, create {AST_ARRAY_TYPE}.make (create {AST_BASIC_TYPE}.make (Ast_bt_char), Void)))
						elseif attached {AST_STRING_CONSTANT} p.item as const and then const.is_eiffel_string then
							-- make sure `is_eiffel_string' is set to `False'
							const.set_is_eiffel_string (False)
						end
					end
				end
			end
		end

	end_visit_unary_expression (a_node: AST_UNARY_EXPRESSION)
		do
			if
				a_node.type = C_op_bitand and
				attached {AST_VARIABLE_EXPRESSION} a_node.expression as ve
			then
				if ve.declaration_binding = printf_declaration then
					extern_printf_needed := True
				end
				if ve.declaration_binding = fprintf_declaration then
					extern_fprintf_needed := True
				end
			end
		end

end
