class
	INTEGER_CONSTANT_EXPRESSION_EVALUATOR

inherit
	ANTLR_AST_TYPE_CONSTANTS

create
	make

feature {NONE}

	make (a_expression: AST_EXPRESSION)
		require
			a_expression /= Void
		local
			l_tmp: like evaluate
		do
			l_tmp := evaluate (a_expression)
			state := l_tmp.state
			evaluated_integer := l_tmp.integer
			evaluated_natural := l_tmp.natural
		end

feature {ANY}

	State_integer,
	State_natural,
	State_no_constant: INTEGER_8 = unique

feature {ANY}

	state: INTEGER_8

	integer: INTEGER_32
		require
			state = State_integer
		do
			Result := evaluated_integer
		end

	natural: NATURAL_32
		require
			state = State_natural
		do
			Result := evaluated_natural
		end

feature {NONE}

	evaluated_integer: INTEGER_32

	evaluated_natural: NATURAL_32

feature {NONE}

	evaluate (a_expression: AST_EXPRESSION): TUPLE [state: INTEGER_8; integer: INTEGER_32; natural: NATURAL_32]
		require
			a_expression /= Void
		local
			l_tmp1, l_tmp2: like evaluate
			l_hex_converter: HEXADECIMAL_STRING_TO_INTEGER_CONVERTER
		do
			if attached {AST_DECIMAL_CONSTANT} a_expression as dc then
				inspect dc.type_binding.type
				when Ast_bt_int, Ast_bt_long then
					Result := [State_integer, dc.value.to_integer_32, {NATURAL_32} 0]
				when Ast_bt_unsigned_int, Ast_bt_unsigned_long then
					Result := [State_natural, 0, dc.value.to_natural_32]
				end
			elseif attached {AST_HEX_CONSTANT} a_expression as hc then
				create l_hex_converter.make
				l_hex_converter.set_leading_separators (" ")
				l_hex_converter.set_trailing_separators (" ")
				l_hex_converter.set_leading_separators_acceptable (True)
				l_hex_converter.set_trailing_separators_acceptable (True)

				inspect hc.type_binding.type
				when Ast_bt_int, Ast_bt_long then
					l_hex_converter.parse_string_with_type (hc.value, {NUMERIC_INFORMATION}.Type_integer_32)
					Result := [State_integer, l_hex_converter.parsed_integer_32, {NATURAL_32} 0]
				when Ast_bt_unsigned_int, Ast_bt_unsigned_long then
					l_hex_converter.parse_string_with_type (hc.value, {NUMERIC_INFORMATION}.Type_natural_32)
					Result := [State_natural, 0, l_hex_converter.parsed_natural_32]
				end
			elseif attached {AST_UNARY_EXPRESSION} a_expression as ue then
				l_tmp1 := evaluate (ue.expression)
				inspect l_tmp1.state
				when State_integer then
					inspect ue.type
					when C_op_plus then  Result := l_tmp1
					when C_op_minus then Result := [State_integer, -l_tmp1.integer, {NATURAL_32} 0]
					end
				when State_natural then
					inspect ue.type
					when C_op_plus then  Result := l_tmp1
					when C_op_minus then Result := [State_natural, 0, (-(l_tmp1.natural.to_integer_32)).to_natural_32]
					end
				when State_no_constant then
					Result := [State_no_constant, 0, {NATURAL_32} 0]
				end
			elseif attached {AST_GENERIC_EXPRESSION} a_expression as ge then
				l_tmp1 := evaluate (ge.left)
				l_tmp2 := evaluate (ge.right)
				if l_tmp1.state = State_integer and l_tmp2.state = State_integer then
					inspect ge.type
					when C_op_plus then        Result := [State_integer, l_tmp1.integer + l_tmp2.integer, {NATURAL_32} 0]
					when C_op_minus then       Result := [State_integer, l_tmp1.integer - l_tmp2.integer, {NATURAL_32} 0]
					when C_op_bitand then      Result := [State_integer, l_tmp1.integer & l_tmp2.integer, {NATURAL_32} 0]
					when C_op_bitor then       Result := [State_integer, l_tmp1.integer | l_tmp2.integer, {NATURAL_32} 0]
					when C_op_exor then        Result := [State_integer, l_tmp1.integer.bit_xor (l_tmp2.integer), {NATURAL_32} 0]
					when C_op_mult then        Result := [State_integer, l_tmp1.integer * l_tmp2.integer, {NATURAL_32} 0]
					when C_op_div then         Result := [State_integer, l_tmp1.integer // l_tmp2.integer, {NATURAL_32} 0]
					when C_op_modulo then      Result := [State_integer, l_tmp1.integer \\ l_tmp2.integer, {NATURAL_32} 0]
					when C_op_left_shift then  Result := [State_integer, l_tmp1.integer |<< l_tmp2.integer, {NATURAL_32} 0]
					when C_op_right_shift then Result := [State_integer, l_tmp1.integer |>> l_tmp2.integer, {NATURAL_32} 0]
					end
				elseif l_tmp1.state = State_natural and l_tmp2.state = State_natural then
					inspect ge.type
					when C_op_plus then   Result := [State_natural, 0, l_tmp1.natural + l_tmp2.natural]
					when C_op_minus then  Result := [State_natural, 0, l_tmp1.natural - l_tmp2.natural]
					when C_op_bitand then Result := [State_natural, 0, l_tmp1.natural & l_tmp2.natural]
					when C_op_bitor then  Result := [State_natural, 0, l_tmp1.natural | l_tmp2.natural]
					when C_op_exor then   Result := [State_natural, 0, l_tmp1.natural.bit_xor (l_tmp2.natural)]
					when C_op_mult then   Result := [State_natural, 0, l_tmp1.natural * l_tmp2.natural]
					when C_op_div then    Result := [State_natural, 0, l_tmp1.natural // l_tmp2.natural]
					when C_op_modulo then Result := [State_natural, 0, l_tmp1.natural \\ l_tmp2.natural]
					end
				elseif l_tmp1.state = State_no_constant or l_tmp2.state = State_no_constant then
					Result := [State_no_constant, 0, {NATURAL_32} 0]
				else
					(create {EXCEPTIONS}).raise ("Integer and Natural constant?!")
				end
			elseif attached {AST_PARENTHESIZED_EXPRESSION} a_expression as pe then
				Result := evaluate (pe.expression)
			elseif
				attached {AST_EIFFEL_EXPRESSION} a_expression as ee and then
				ee.mixed.count = 2 and then
				attached {AST_EXPRESSION} ee.mixed.first as e and then
				attached {STRING_8} ee.mixed.last as s
			then
				l_tmp1 := evaluate (e)
				inspect l_tmp1.state
				when State_integer then
					    if s.is_equal (".to_integer_32") then Result := l_tmp1
					elseif s.is_equal (".to_natural_32") then Result := [State_natural, 0, l_tmp1.integer.to_natural_32]
					else
						(create {EXCEPTIONS}).raise ("Unsupported cast: " + a_expression.c_code)
					end
				when State_natural then
					    if s.is_equal (".to_integer_32") then Result := [State_integer, l_tmp1.natural.to_integer_32, {NATURAL_32} 0]
					elseif s.is_equal (".to_natural_32") then Result := l_tmp1
					else
						(create {EXCEPTIONS}).raise ("Unsupported cast: " + a_expression.c_code)
					end
				when State_no_constant then
					Result := [State_no_constant, 0, {NATURAL_32} 0]
				end
			else
				Result := [State_no_constant, 0, {NATURAL_32} 0]
			end
		ensure
			Result /= Void
			Result.state = State_integer or Result.state = State_natural or Result.state = State_no_constant
		end

end
